from __future__ import annotations  # noqa: F401

import typing
from collections import defaultdict

import datatree as dt
import xarray as xr

from .common import Projection
from .utils import add_metadata_and_zarr_encoding, get_version, multiscales_template


def pyramid_coarsen(
    ds: xr.Dataset, *, factors: list[int], dims: list[str], **kwargs
) -> dt.DataTree:
    """Create a multiscale pyramid via coarsening of a dataset by given factors

    Parameters
    ----------
    ds : xarray.Dataset
        The dataset to coarsen.
    factors : list[int]
        The factors to coarsen by.
    dims : list[str]
        The dimensions to coarsen.
    kwargs : dict
        Additional keyword arguments to pass to xarray.Dataset.coarsen.
    """

    # multiscales spec
    save_kwargs = locals()
    del save_kwargs['ds']

    attrs = {
        'multiscales': multiscales_template(
            datasets=[{'path': str(i)} for i in range(len(factors))],
            type='reduce',
            method='pyramid_coarsen',
            version=get_version(),
            kwargs=save_kwargs,
        )
    }

    # set up pyramid
    plevels = {}

    # pyramid data
    for key, factor in enumerate(factors):
        # merge dictionary via union operator
        kwargs |= {d: factor for d in dims}
        plevels[str(key)] = ds.coarsen(**kwargs).mean()

    plevels['/'] = xr.Dataset(attrs=attrs)
    return dt.DataTree.from_dict(plevels)


def pyramid_reproject(
    ds: xr.Dataset,
    *,
    projection:typing.Literal['web-mercator', 'equidistant-cylindrical'] = 'web-mercator',
    levels: int = None,
    pixels_per_tile: int = 128,
    other_chunks: dict = None,
    resampling: str | dict = 'average',
    extra_dim: str = None,
) -> dt.DataTree:

    """Create a multiscale pyramid of a dataset via reprojection.

    Parameters
    ----------
    ds : xarray.Dataset
        The dataset to create a multiscale pyramid of.
    projection : str, optional
        The projection to use. Default is 'web-mercator'.
    levels : int, optional
        The number of levels to create. If None, the number of levels is
        determined by the number of tiles in the dataset.
    pixels_per_tile : int, optional
        Number of pixels per tile, by default 128
    other_chunks : dict
        Chunks for non-spatial dims to pass to :py:meth:`~xr.Dataset.chunk`. Default is None
    resampling : str or dict, optional
        Rasterio warp resampling method to use. Default is 'average'.
        If a dict, keys are variable names and values are warp resampling methods.
    extra_dim : str, optional
        The name of the extra dimension to iterate over. Default is None.

    Returns
    -------
    dt.DataTree
        The multiscale pyramid.

    """

    import rioxarray  # noqa: F401
    from rasterio.warp import Resampling

    # multiscales spec
    save_kwargs = {'levels': levels, 'pixels_per_tile': pixels_per_tile}
    attrs = {
        'multiscales': multiscales_template(
            datasets=[{'path': str(i)} for i in range(levels)],
            type='reduce',
            method='pyramid_reproject',
            version=get_version(),
            kwargs=save_kwargs,
        )
    }

    # Convert resampling from string to dictionary if necessary
    if isinstance(resampling, str):
        resampling_dict = defaultdict(lambda: resampling)
    else:
        resampling_dict = resampling

    projection_model = Projection(name=projection)

    # set up pyramid
    plevels = {}

    # pyramid data
    for level in range(levels):
        lkey = str(level)
        dim = 2**level * pixels_per_tile
        dst_transform = projection_model.transform(dim=dim)

        def reproject(da, var):
            return da.rio.reproject(
                projection_model._crs,
                resampling=Resampling[resampling_dict[var]],
                shape=(dim, dim),
                transform=dst_transform,
            )

        # create the data array for each level
        plevels[lkey] = xr.Dataset(attrs=ds.attrs)
        for k, da in ds.items():
            if len(da.shape) == 4:
                # if extra_dim is not specified, raise an error
                if extra_dim is None:
                    raise ValueError("must specify 'extra_dim' to iterate over 4d data")
                da_all = []
                for index in ds[extra_dim]:
                    # reproject each index of the 4th dimension
                    da_reprojected = reproject(da.sel({extra_dim: index}), k)
                    da_all.append(da_reprojected)
                plevels[lkey][k] = xr.concat(da_all, ds[extra_dim])
            else:
                # if the data array is not 4D, just reproject it
                plevels[lkey][k] = reproject(da, k)

    # create the final multiscale pyramid
    plevels['/'] = xr.Dataset(attrs=attrs)
    pyramid = dt.DataTree.from_dict(plevels)

    pyramid = add_metadata_and_zarr_encoding(
        pyramid, levels=levels, pixels_per_tile=pixels_per_tile, other_chunks=other_chunks, projection=projection_model
    )
    return pyramid



import pytest

import ndpyramid


@pytest.mark.parametrize(
    'datasets,type,method,version,args,kwargs', [(None, '', '', '', None, None)]
)
def test_multiscales_template(datasets, type, method, version, args, kwargs):
    template = ndpyramid.utils.multiscales_template(
        datasets=datasets, type=type, method=method, version=version, args=args, kwargs=kwargs
    )[0]
    if not kwargs:
        assert template['metadata']['kwargs'] == {}
    if not datasets:
        assert template['datasets'] == []
    if not args:
        assert template['metadata']['args'] == []
    assert template['type'] == type
    assert template['metadata']['method'] == method
    assert template['metadata']['version'] == version



import typing

import pydantic
import pyproj
import rasterio.transform


class Projection(pydantic.BaseModel):
    name: typing.Literal['web-mercator', 'equidistant-cylindrical'] = 'web-mercator'
    _crs: str = pydantic.PrivateAttr()
    _proj = pydantic.PrivateAttr()

    def __init__(self, **data) -> None:

        super().__init__(**data)
        epsg_codes = {'web-mercator': 'EPSG:3857', 'equidistant-cylindrical': 'EPSG:4326'}
        self._crs = epsg_codes[self.name]
        self._proj = pyproj.Proj(self._crs)

    @pydantic.validate_arguments
    def transform(self, *, dim:int) -> rasterio.transform.Affine:


        if self.name == 'web-mercator':
            # set up the transformation matrix for the web-mercator projection such that the data conform
            # to the slippy-map tiles assumed boundaries. See https://github.com/carbonplan/ndpyramid/pull/70
            # for detailed on calculating the parameters.
            return rasterio.transform.Affine.translation(-20037508.342789244, 20037508.342789248) * rasterio.transform.Affine.scale((20037508.342789244 * 2) / dim, -(20037508.342789248 * 2) / dim)
        elif self.name == 'equidistant-cylindrical':
            # set up the transformation matrix that maps between the Equidistant Cylindrical projection
            # and the latitude-longitude projection. The Affine.translation function moves the origin
            # of the grid from (0, 0) to (-180, 90) in latitude-longitude coordinates,
            # and the Affine.scale function scales the grid coordinates to match the size of the grid
            # in latitude-longitude coordinates. The resulting transformation matrix maps grid coordinates to
            # latitude-longitude coordinates.
            return rasterio.transform.Affine.translation(-180, 90) * rasterio.transform.Affine.scale(360 / dim, -180 / dim)



import numpy as np
import pytest
import xarray as xr
from zarr.storage import MemoryStore

from ndpyramid import pyramid_coarsen, pyramid_regrid, pyramid_reproject
from ndpyramid.regrid import generate_weights_pyramid, make_grid_ds


@pytest.fixture
def temperature():
    ds = xr.tutorial.open_dataset('air_temperature')
    ds['air'].encoding = {}
    return ds


def test_xarray_coarsened_pyramid(temperature):
    factors = [4, 2, 1]
    pyramid = pyramid_coarsen(temperature, dims=('lat', 'lon'), factors=factors, boundary='trim')
    assert pyramid.ds.attrs['multiscales']
    assert len(pyramid.ds.attrs['multiscales'][0]['datasets']) == len(factors)
    pyramid.to_zarr(MemoryStore())


def test_reprojected_pyramid(temperature):
    pytest.importorskip('rioxarray')
    levels = 2
    temperature = temperature.rio.write_crs('EPSG:4326')
    pyramid = pyramid_reproject(temperature, levels=2)
    assert pyramid.ds.attrs['multiscales']
    assert len(pyramid.ds.attrs['multiscales'][0]['datasets']) == levels
    assert pyramid.ds.attrs['multiscales'][0]['datasets'][0]['crs'] == 'EPSG:3857'
    pyramid.to_zarr(MemoryStore())


@pytest.mark.parametrize('regridder_apply_kws', [None, {'keep_attrs': False}])
def test_regridded_pyramid(temperature, regridder_apply_kws):
    pytest.importorskip('xesmf')
    pyramid = pyramid_regrid(
        temperature, levels=2, regridder_apply_kws=regridder_apply_kws, other_chunks={'time': 2}
    )
    assert pyramid.ds.attrs['multiscales']
    expected_attrs = (
        temperature['air'].attrs
        if not regridder_apply_kws or regridder_apply_kws.get('keep_attrs')
        else {}
    )
    assert pyramid['0'].ds.air.attrs == expected_attrs
    assert pyramid['1'].ds.air.attrs == expected_attrs
    pyramid.to_zarr(MemoryStore())


def test_regridded_pyramid_with_weights(temperature):
    pytest.importorskip('xesmf')
    levels = 2
    weights_pyramid = generate_weights_pyramid(temperature.isel(time=0), levels)
    pyramid = pyramid_regrid(
        temperature, levels=levels, weights_pyramid=weights_pyramid, other_chunks={'time': 2}
    )
    assert pyramid.ds.attrs['multiscales']
    assert len(pyramid.ds.attrs['multiscales'][0]['datasets']) == levels
    pyramid.to_zarr(MemoryStore())


@pytest.mark.parametrize('projection', ['web-mercator', 'equidistant-cylindrical'])
def test_make_grid_ds(projection):

    grid = make_grid_ds(0, pixels_per_tile=8, projection=projection)
    lon_vals = grid.lon_b.values
    assert np.all((lon_vals[-1, :] - lon_vals[0, :]) < 0.001)
    assert grid.attrs['title'] == 'Web Mercator Grid' if projection == 'web-mercator' else 'Equidistant Cylindrical Grid'

@pytest.mark.parametrize('levels', [1, 2])
@pytest.mark.parametrize('method', ['bilinear', 'conservative'])
def test_generate_weights_pyramid(temperature, levels, method):
    weights_pyramid = generate_weights_pyramid(temperature.isel(time=0), levels, method=method)
    assert weights_pyramid.ds.attrs['levels'] == levels
    assert weights_pyramid.ds.attrs['regrid_method'] == method
    assert set(weights_pyramid['0'].ds.data_vars) == {'S', 'col', 'row'}
    assert 'n_in' in weights_pyramid['0'].ds.attrs and 'n_out' in weights_pyramid['0'].ds.attrs



from __future__ import annotations  # noqa: F401

import itertools
import typing

import datatree as dt
import numpy as np
import xarray as xr

from .common import Projection
from .utils import add_metadata_and_zarr_encoding, get_version, multiscales_template


def xesmf_weights_to_xarray(regridder) -> xr.Dataset:
    w = regridder.weights.data
    dim = 'n_s'
    ds = xr.Dataset(
        {
            'S': (dim, w.data),
            'col': (dim, w.coords[1, :] + 1),
            'row': (dim, w.coords[0, :] + 1),
        }
    )
    ds.attrs = {'n_in': regridder.n_in, 'n_out': regridder.n_out}
    return ds


def _reconstruct_xesmf_weights(ds_w):
    """Reconstruct weights into format that xESMF understands"""
    import sparse
    import xarray as xr

    col = ds_w['col'].values - 1
    row = ds_w['row'].values - 1
    s = ds_w['S'].values
    n_out, n_in = ds_w.attrs['n_out'], ds_w.attrs['n_in']
    crds = np.stack([row, col])
    return xr.DataArray(
        sparse.COO(crds, s, (n_out, n_in)), dims=('out_dim', 'in_dim'), name='weights'
    )


def make_grid_ds(level: int, pixels_per_tile: int = 128, projection:typing.Literal['web-mercator', 'equidistant-cylindrical'] = 'web-mercator') -> xr.Dataset:
    """Make a dataset representing a target grid

    Parameters
    ----------
    level : int
        The zoom level to compute the grid for. Level zero is the furthest out zoom level
    pixels_per_tile : int, optional
        Number of pixels to include along each axis in individual tiles, by default 128
    projection : str, optional
        The projection to use for the grid, by default 'equidistant-cylindrical'

    Returns
    -------
    xr.Dataset
        Target grid dataset with the following variables:
        - "x": X coordinate in Web Mercator projection (grid cell center)
        - "y": Y coordinate in Web Mercator projection (grid cell center)
        - "lat": latitude coordinate (grid cell center)
        - "lon": longitude coordinate (grid cell center)
        - "lat_b": latitude bounds for grid cell
        - "lon_b": longitude bounds for grid cell
    """

    projection_model = Projection(name=projection)

    dim = (2**level) * pixels_per_tile

    transform = projection_model.transform(dim=dim)

    if projection_model.name == 'equidistant-cylindrical':
        title='Equidistant Cylindrical Grid'

    elif projection_model.name == 'web-mercator':
        title='Web Mercator Grid'

    p = projection_model._proj


    grid_shape = (dim, dim)
    bounds_shape = (dim + 1, dim + 1)

    xs = np.empty(grid_shape)
    ys = np.empty(grid_shape)
    lat = np.empty(grid_shape)
    lon = np.empty(grid_shape)
    lat_b = np.zeros(bounds_shape)
    lon_b = np.zeros(bounds_shape)

    # calc grid cell center coordinates
    ii, jj = np.meshgrid(np.arange(dim) + 0.5, np.arange(dim) + 0.5)
    for i, j in itertools.product(range(grid_shape[0]), range(grid_shape[1])):
        locs = [ii[i, j], jj[i, j]]
        xs[i, j], ys[i, j] = transform * locs
        lon[i, j], lat[i, j] = p(xs[i, j], ys[i, j], inverse=True)

    # calc grid cell bounds
    iib, jjb = np.meshgrid(np.arange(dim + 1), np.arange(dim + 1))
    for i, j in itertools.product(range(bounds_shape[0]), range(bounds_shape[1])):
        locs = [iib[i, j], jjb[i, j]]
        x, y = transform * locs
        lon_b[i, j], lat_b[i, j] = p(x, y, inverse=True)

    return xr.Dataset(
        {
            'x': xr.DataArray(xs[0, :], dims=['x']),
            'y': xr.DataArray(ys[:, 0], dims=['y']),
            'lat': xr.DataArray(lat, dims=['y', 'x']),
            'lon': xr.DataArray(lon, dims=['y', 'x']),
            'lat_b': xr.DataArray(lat_b, dims=['y_b', 'x_b']),
            'lon_b': xr.DataArray(lon_b, dims=['y_b', 'x_b']),
        },
        attrs=dict(title=title, Conventions='CF-1.8'),
    )


def make_grid_pyramid(levels: int = 6, projection:typing.Literal['web-mercator', 'equidistant-cylindrical'] = 'web-mercator') -> dt.DataTree:
    """helper function to create a grid pyramid for use with xesmf

    Parameters
    ----------
    levels : int, optional
        Number of levels in pyramid, by default 6

    Returns
    -------
    pyramid : dt.DataTree
        Multiscale grid definition
    """
    plevels = {
        str(level): make_grid_ds(level, projection=projection).chunk(-1) for level in range(levels)
    }
    return dt.DataTree.from_dict(plevels)


def generate_weights_pyramid(
    ds_in: xr.Dataset, levels: int, method: str = 'bilinear', regridder_kws: dict = None,
    projection:typing.Literal['web-mercator', 'equidistant-cylindrical'] = 'web-mercator'

) -> dt.DataTree:
    """helper function to generate weights for a multiscale regridder

    Parameters
    ----------
    ds_in : xr.Dataset
        Input dataset to regrid
    levels : int
        Number of levels in the pyramid
    method : str, optional
        Regridding method. See :py:class:`~xesmf.Regridder` for valid options, by default 'bilinear'
    regridder_kws : dict
        Keyword arguments to pass to :py:class:`~xesmf.Regridder`. Default is `{'periodic': True}`
    projection : str, optional
        The projection to use for the grid, by default 'web-mercator'

    Returns
    -------
    weights : dt.DataTree
        Multiscale weights
    """
    import xesmf as xe

    regridder_kws = {} if regridder_kws is None else regridder_kws
    regridder_kws = {'periodic': True, **regridder_kws}

    plevels = {}
    for level in range(levels):
        ds_out = make_grid_ds(level=level, projection=projection)
        regridder = xe.Regridder(ds_in, ds_out, method, **regridder_kws)
        ds = xesmf_weights_to_xarray(regridder)

        plevels[str(level)] = ds

    root = xr.Dataset(attrs={'levels': levels, 'regrid_method': method})
    plevels['/'] = root
    return dt.DataTree.from_dict(plevels)


def pyramid_regrid(
    ds: xr.Dataset,
    projection:typing.Literal['web-mercator', 'equidistant-cylindrical'] = 'web-mercator',
    target_pyramid: dt.DataTree = None,
    levels: int = None,
    weights_pyramid: dt.DataTree = None,
    method: str = 'bilinear',
    regridder_kws: dict = None,
    regridder_apply_kws: dict = None,
    other_chunks: dict = None,
    pixels_per_tile: int = 128,

) -> dt.DataTree:
    """Make a pyramid using xesmf's regridders

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset
    projection : str, optional
        Projection to use for the grid, by default 'web-mercator'
    target_pyramid : dt.DataTree, optional
        Target grids, if not provided, they will be generated, by default None
    levels : int, optional
        Number of levels in pyramid, by default None
    weights_pyramid : dt.DataTree, optional
       pyramid containing pregenerated weights
    method : str, optional
        Regridding method. See :py:class:`~xesmf.Regridder` for valid options, by default 'bilinear'
    regridder_kws : dict
        Keyword arguments to pass to regridder. Default is `{'periodic': True}`
    regridder_apply_kws : dict
        Keyword arguments such as `keep_attrs`, `skipna`, `na_thres`
        to pass to :py:meth:`~xesmf.Regridder.__call__`. Default is None
    other_chunks : dict
        Chunks for non-spatial dims to pass to :py:meth:`~xr.Dataset.chunk`. Default is None
    pixels_per_tile : int, optional
        Number of pixels per tile, by default 128

    Returns
    -------
    pyramid : dt.DataTree
        Multiscale data pyramid
    """
    import xesmf as xe

    if target_pyramid is None:
        if levels is not None:
            target_pyramid = make_grid_pyramid(levels, projection=projection)
        else:
            raise ValueError('must either provide a target_pyramid or number of levels')
    if levels is None:
        levels = len(target_pyramid.keys())  # TODO: get levels from the pyramid metadata

    regridder_kws = {} if regridder_kws is None else regridder_kws
    regridder_kws = {'periodic': True, **regridder_kws}

    # multiscales spec
    save_kwargs = locals()
    del save_kwargs['ds']
    del save_kwargs['target_pyramid']
    del save_kwargs['xe']
    del save_kwargs['weights_pyramid']

    attrs = {
        'multiscales': multiscales_template(
            datasets=[{'path': str(i)} for i in range(levels)],
            type='reduce',
            method='pyramid_regrid',
            version=get_version(),
            kwargs=save_kwargs,
        )
    }

    # set up pyramid

    plevels = {}

    # pyramid data
    for level in range(levels):
        grid = target_pyramid[str(level)].ds.load()
        # get the regridder object
        if weights_pyramid is None:
            regridder = xe.Regridder(ds, grid, method, **regridder_kws)
        else:
            # Reconstruct weights into format that xESMF understands
            # this is a hack that assumes the weights were generated by
            # the `generate_weights_pyramid` function

            ds_w = weights_pyramid[str(level)].ds
            weights = _reconstruct_xesmf_weights(ds_w)
            regridder = xe.Regridder(
                ds, grid, method, reuse_weights=True, weights=weights, **regridder_kws
            )
        # regrid
        if regridder_apply_kws is None:
            regridder_apply_kws = {}
        regridder_apply_kws = {**{'keep_attrs': True}, **regridder_apply_kws}
        plevels[str(level)] = regridder(ds, **regridder_apply_kws)

    root = xr.Dataset(attrs=attrs)
    plevels['/'] = root
    pyramid = dt.DataTree.from_dict(plevels)

    pyramid = add_metadata_and_zarr_encoding(
        pyramid, levels=levels, other_chunks=other_chunks, pixels_per_tile=pixels_per_tile, projection=Projection(name=projection)
    )

    return pyramid



from __future__ import annotations  # noqa: F401

import contextlib

import cf_xarray  # noqa: F401
import datatree as dt
import numpy as np
import numpy.typing as npt
import xarray as xr

from ._version import __version__
from .common import Projection

# from netCDF4 and netCDF4-python
default_fillvals = {
    'S1': '\x00',
    'i1': -127,
    'u1': 255,
    'i2': -32767,
    'u2': 65535,
    'i4': -2147483647,
    'u4': 4294967295,
    'i8': -9223372036854775806,
    'u8': 18446744073709551614,
    'f4': 9.969209968386869e36,
    'f8': 9.969209968386869e36,
}


def get_version() -> str:
    return __version__


def multiscales_template(
    *,
    datasets: list = None,
    type: str = '',
    method: str = '',
    version: str = '',
    args: list = None,
    kwargs: dict = None,
):
    if datasets is None:
        datasets = []
    if args is None:
        args = []
    if kwargs is None:
        kwargs = {}
    # https://forum.image.sc/t/multiscale-arrays-v0-1/37930
    return [
        {
            'datasets': datasets,
            'type': type,
            'metadata': {'method': method, 'version': version, 'args': args, 'kwargs': kwargs},
        }
    ]


def set_zarr_encoding(
    ds: xr.Dataset,
    codec_config: dict | None = None,
    float_dtype: npt.DTypeLike | None = None,
    int_dtype: npt.DTypeLike | None = None,
) -> xr.Dataset:
    """Set zarr encoding for each variable in the dataset

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset
    codec_config : dict, optional
        Dictionary of parameters to pass to numcodecs.get_codec.
        The default is {'id': 'zlib', 'level': 1}
    float_dtype : str or dtype, optional
        Dtype to cast floating point variables to

    Returns
    -------
    ds : xr.Dataset
        Output dataset with updated variable encodings
    """
    import numcodecs

    ds = ds.copy()

    if codec_config is None:
        codec_config = {'id': 'zlib', 'level': 1}
    compressor = numcodecs.get_codec(codec_config)

    time_vars = ds.cf.axes.get('T', []) + ds.cf.bounds.get('T', [])
    for varname, da in ds.variables.items():
        # maybe cast float type
        if np.issubdtype(da.dtype, np.floating) and float_dtype is not None:
            da = da.astype(float_dtype)

        if np.issubdtype(da.dtype, np.integer) and int_dtype is not None:
            da = da.astype(int_dtype)

        # remove old encoding
        da.encoding.clear()

        # update with new encoding
        da.encoding['compressor'] = compressor
        with contextlib.suppress(KeyError):
            del da.attrs['_FillValue']
        da.encoding['_FillValue'] = default_fillvals.get(da.dtype.str[-2:], None)

        # TODO: handle date/time types
        # set encoding for time and time_bnds
        if varname in time_vars:
            da.encoding['dtype'] = 'int32'
        ds[varname] = da

    return ds


def add_metadata_and_zarr_encoding(
    pyramid: dt.DataTree, *, levels: int, other_chunks: dict = None, pixels_per_tile: int = 128, projection: Projection = None
) -> dt.DataTree:

    '''Postprocess data pyramid. Adds multiscales metadata and sets Zarr encoding

    Parameters
    ----------
    pyramid : dt.DataTree
        Input data pyramid
    levels : int
        Number of levels in pyramid
    other_chunks : dict
        Chunks for non-spatial dims
    pixels_per_tile : int
        Number of pixels per tile
    projection: Projection
        Projection model of the pyramids

    Returns
    -------
    dt.DataTree
        Updated data pyramid with metadata / encoding set
    '''
    chunks = {'x': pixels_per_tile, 'y': pixels_per_tile}
    if other_chunks is not None:
        chunks |= other_chunks

    for level in range(levels):
        slevel = str(level)
        pyramid.ds.attrs['multiscales'][0]['datasets'][level]['pixels_per_tile'] = pixels_per_tile
        if projection:
            pyramid.ds.attrs['multiscales'][0]['datasets'][level]['crs'] = projection._crs
        # set dataset chunks
        pyramid[slevel].ds = pyramid[slevel].ds.chunk(chunks)

        # set dataset encoding
        pyramid[slevel].ds = set_zarr_encoding(
            pyramid[slevel].ds, codec_config={'id': 'zlib', 'level': 1}, float_dtype='float32'
        )

    # set global metadata
    pyramid.ds.attrs.update({'title': 'multiscale data pyramid', 'version': __version__})
    return pyramid
